// Netify Agent Device Discovery Processor
// Copyright (C) 2023 eGloo Incorporated
// <http://www.egloo.ca>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#pragma once

#include <condition_variable>
#include <deque>
#include <map>
#include <mutex>
#include <queue>
#include <set>
#include <string>

#include <nd-napi.hpp>
#include <nd-plugin.hpp>

#define _NPP_MAX_CONFIDENCE    80
#define _NPP_MAX_HTTP_UA       10
#define _NPP_MAX_MDNS_SERVICES 10

class nppGlobalConfig
{
public:
    size_t max_confidence;
    size_t max_http_user_agents;
    size_t max_mdns_services;

    bool api_enabled;
    string api_url, api_key;

    nppGlobalConfig() { Reset(); }

    void Reset(void) {
        max_confidence = _NPP_MAX_CONFIDENCE;
        max_http_user_agents = _NPP_MAX_HTTP_UA;
        max_mdns_services = _NPP_MAX_MDNS_SERVICES;

        api_enabled = false;

        api_url.clear();
        api_key.clear();
    }

    void Load(const json &j);
};

class nppChannelConfig
{
public:
    enum Format {
        FORMAT_RAW,
        FORMAT_JSON,
        FORMAT_MSGPACK,
    };

    enum Compressor {
        COMPRESSOR_NONE,
        COMPRESSOR_GZ,
    };

    nppChannelConfig()
      : format(FORMAT_JSON), compressor(COMPRESSOR_NONE) { }

    void Load(const string &channel, const json &jconf);
    inline void Load(const string &channel,
      const json &jconf, nppChannelConfig &defaults) {
        format = defaults.format;
        compressor = defaults.compressor;
        Load(channel, jconf);
    }

    string channel;
    Format format;
    Compressor compressor;
};

class nppDevice
{
public:
    nppDevice() : updates(1), confidence(0) { }

    void Update(const nppGlobalConfig &config,
      const nppDevice &device);

    enum EncodeFlags {
        ET_DHCP = 0x01,
        ET_MDNS = 0x02,
        ET_HTTP = 0x04,
        ET_CONFIDENCE = 0x08,
        ET_IDENTITY = 0x10,
        ET_METADATA = (0x01 | 0x02 | 0x04),
    };

    void Encode(json &j, uint8_t flags = ET_METADATA);

    string mac;
    string label;
    string dhcp_fingerprint;
    string dhcp_vendor;
    string dhcp_hostname;

    typedef deque<string> UniqueSet;

    UniqueSet mdns_services;
    UniqueSet http_user_agents;

    unsigned updates;
    unsigned confidence;
};

class nppPlugin;

class nppApiDeviceDiscovery : public ndNetifyApiThread
{
public:
    nppApiDeviceDiscovery(const nppGlobalConfig &config)
      : config(config), method(ndNetifyApiThread::METHOD_GET) { }

    nppApiDeviceDiscovery(const nppGlobalConfig &config,
      const json &json)
      : config(config), method(ndNetifyApiThread::METHOD_POST) {

        payload = json.dump(ndGC_DEBUG ? ND_JSON_INDENT : -1,
          ' ', true, json::error_handler_t::replace);
        mime_type = "application/json";
    };

    virtual void *Entry(void);

protected:
    friend class nppPlugin;

    const nppGlobalConfig &config;
    ndNetifyApiThread::Method method;
    string payload;
    string mime_type;
};

class nppPlugin : public ndPluginProcessor
{
public:
    nppPlugin(const string &tag, const ndPlugin::Params &params);
    virtual ~nppPlugin();

    virtual void *Entry(void);

    virtual void GetVersion(string &version) {
        version = PACKAGE_VERSION;
    }
    template <class T>
    void GetStatus(T &output) const {
        nd_dprintf("%s: TODO\n", __PRETTY_FUNCTION__);
        ndPluginProcessor::GetStatus(output);
    }

    virtual void DispatchEvent(ndPlugin::Event event,
      void *param = nullptr);

    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndFlowMap *flow_map);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, nd_flow_ptr &flow);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndInterfaces *interfaces);
    virtual void DispatchProcessorEvent(ndPluginProcessor::Event event,
      const string &iface, ndPacketStats *stats);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndPacketStats *stats);
    virtual void DispatchProcessorEvent(
      ndPluginProcessor::Event event, ndInstanceStatus *status);
    virtual void
    DispatchProcessorEvent(ndPluginProcessor::Event event);

protected:
    atomic<bool> reload;
    atomic<bool> update;

    void Reload(void);

    nppGlobalConfig globals;
    nppChannelConfig defaults;

    typedef map<string, map<string, nppChannelConfig>> Sinks;
    Sinks sinks;

    typedef map<string, nppDevice> Devices;
    Devices devices;

    typedef queue<nd_flow_ptr> FlowQueue;
    FlowQueue flow_queue;

    mutex flow_queue_mutex;
    condition_variable flow_queue_cond;

    typedef set<string> Identified;
    Identified identified;

    void ProcessFlow(nd_flow_ptr &flow);
    void PrintDeviceDebug(const nppDevice &device) const;
    void ProcessUpdate(void);

    void DispatchPayload(json &j);

    nppApiDeviceDiscovery *api_thread;

    void ProcessApiResponse(void);
};
